@file:JvmName("TransactionMethodsKt")

package ru.landgrafhomyak.serdha.api.v0.runtime

import kotlin.contracts.InvocationKind
import kotlin.contracts.contract
import kotlin.jvm.JvmName
import ru.landgrafhomyak.serdha.api.v0.LowLevelApi
import ru.landgrafhomyak.serdha.api.v0.dml.Delete
import ru.landgrafhomyak.serdha.api.v0.dml.Insert
import ru.landgrafhomyak.serdha.api.v0.dml.InsertCreator
import ru.landgrafhomyak.serdha.api.v0.dml.Select
import ru.landgrafhomyak.serdha.api.v0.dml.Update


@Suppress("FunctionName")
@PublishedApi
internal inline fun <R> _safeAutoClose(onAbort: () -> Unit = {}, onSuccess: () -> Unit = {}, action: () -> R): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(onAbort, InvocationKind.AT_MOST_ONCE)
		callsInPlace(onSuccess, InvocationKind.AT_MOST_ONCE)
	}
	val ret: R
	try {
		ret = action()
	} catch (e1: Throwable) {
		try {
			onAbort()
		} catch (e2: Throwable) {
			e1.addSuppressed(e2)
		}
		throw e1
	}
	onSuccess()
	return ret
}

@Suppress("FunctionName")
@PublishedApi
internal inline fun <R> _safeAutoClose(finally: () -> Unit, action: () -> R): R {
	contract {
		callsInPlace(action, InvocationKind.EXACTLY_ONCE)
		callsInPlace(finally, InvocationKind.EXACTLY_ONCE)
	}
	return _safeAutoClose(finally, finally, action)
}

public class QueryWithoutReturnError(message: String, public val query: Any) : Error(message)

@Suppress("FunctionName", "DuplicatedCode")
@PublishedApi
@OptIn(LowLevelApi::class)
internal suspend inline fun <Q : Any> _wrapWithLambdas(
	compiledQuery: Any,
	queryGetter: () -> _ParametersSetter<Q, _ResultSet<Q, Unit>?>,
	parameters: (ParametersSetter<Q>) -> Unit,
	hasReturning: Boolean,
	returning: (_ResultSet<Q, *>) -> Unit,
	queryWithoutErrorMessage: String
) {
	contract {
		callsInPlace(queryGetter, InvocationKind.EXACTLY_ONCE)
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(returning, InvocationKind.AT_MOST_ONCE)
	}
	val paramsSetter = queryGetter()

	_safeAutoClose(onAbort = { paramsSetter._abort() }) {
		parameters(paramsSetter)
	}

	val rows = paramsSetter._execute()
	if (rows == null) {
		if (hasReturning)
			throw QueryWithoutReturnError(queryWithoutErrorMessage, compiledQuery)
		return
	}

	_safeAutoClose(onAbort = { rows._abort() }, onSuccess = { rows._close() }) {
		returning(rows)
	}
}

public class ResultNotSingleError : Error()

@Suppress("FunctionName")
@PublishedApi
@OptIn(LowLevelApi::class)
internal suspend inline fun <Q : Any, R> _wrapWithLambdasSingleOrNull(
	compiledQuery: Any,
	queryGetter: () -> _ParametersSetter<Q, _ResultSet<Q, Unit>?>,
	parameters: (ParametersSetter<Q>) -> Unit,
	hasReturning: Boolean,
	returning: (Row<Q>) -> R,
	queryWithoutErrorMessage: String
): R? {
	contract {
		callsInPlace(queryGetter, InvocationKind.EXACTLY_ONCE)
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(returning, InvocationKind.AT_MOST_ONCE)
	}

	_wrapWithLambdas(
		compiledQuery, queryGetter, parameters, hasReturning,
		{ rs ->
			if (!rs._next())
				return@_wrapWithLambdasSingleOrNull null
			val ret = returning(rs)
			if (rs._next())
				throw ResultNotSingleError()
			return@_wrapWithLambdasSingleOrNull ret
		},
		queryWithoutErrorMessage
	)
	throw QueryWithoutReturnError(queryWithoutErrorMessage, compiledQuery)
}


@Suppress("FunctionName")
@PublishedApi
@OptIn(LowLevelApi::class)
internal suspend inline fun <Q : Any> _wrapWithLambdasIterate(
	compiledQuery: Any,
	queryGetter: () -> _ParametersSetter<Q, _ResultSet<Q, Unit>?>,
	parameters: (ParametersSetter<Q>) -> Unit,
	hasReturning: Boolean,
	returning: (Row<Q>) -> Unit,
	queryWithoutErrorMessage: String
) {
	contract {
		callsInPlace(queryGetter, InvocationKind.EXACTLY_ONCE)
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(returning, InvocationKind.UNKNOWN)
	}

	return _wrapWithLambdas(compiledQuery, queryGetter, parameters, hasReturning, { rs -> while (rs._next()) returning(rs) }, queryWithoutErrorMessage)
}

@Suppress("FunctionName")
@PublishedApi
@OptIn(LowLevelApi::class)
internal suspend inline fun <Q : Any, R> _wrapWithLambdasMap(
	compiledQuery: Any,
	queryGetter: () -> _ParametersSetter<Q, _ResultSet<Q, Unit>?>,
	parameters: (ParametersSetter<Q>) -> Unit,
	hasReturning: Boolean,
	returning: (Row<Q>) -> R,
	queryWithoutErrorMessage: String
): List<R> {
	contract {
		callsInPlace(queryGetter, InvocationKind.EXACTLY_ONCE)
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(returning, InvocationKind.UNKNOWN)
	}

	_wrapWithLambdas(
		compiledQuery, queryGetter, parameters, hasReturning,
		{ rs ->
			val out = ArrayList<R>()
			while (rs._next())
				out.add(returning(rs))
			return@_wrapWithLambdasMap out
		},
		queryWithoutErrorMessage
	)
	throw QueryWithoutReturnError(queryWithoutErrorMessage, compiledQuery)
}


@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> SynchronizedDatabase<*>.autoTransactedSelectSingleOrNull(
	compiledQuery: Select<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.AT_MOST_ONCE)
	}
	return _wrapWithLambdasSingleOrNull(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedSelect(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.selectSingleOrNull(
	compiledQuery: Select<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.AT_MOST_ONCE)
	}
	return _wrapWithLambdasSingleOrNull(
		compiledQuery = compiledQuery,
		queryGetter = { this._select(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}


@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.selectThenIterate(
	compiledQuery: Select<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	_wrapWithLambdasIterate(
		compiledQuery = compiledQuery,
		queryGetter = { this._select(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> SynchronizedDatabase<*>.autoTransactedSelectThenIterate(
	compiledQuery: Select<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	_wrapWithLambdasIterate(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedSelect(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.selectThenMap(
	compiledQuery: Select<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): List<R> {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	return _wrapWithLambdasMap(
		compiledQuery = compiledQuery,
		queryGetter = { this._select(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}


@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> SynchronizedDatabase<*>.autoTransactedSelectThenMap(
	compiledQuery: Select<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): List<R> {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	return _wrapWithLambdasMap(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedSelect(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@Suppress("FunctionName", "DuplicatedCode")
@PublishedApi
@OptIn(LowLevelApi::class)
internal suspend inline fun <Q : Any, T> _wrapInsertParamsWithLambdas(
	compiledQuery: Insert.InsertParams<Q>,
	queryGetter: () -> _ParametersSetter<Q, _ParametersSetter._Multi<InsertCreator.InsertParams.DataParam<Q>, _ResultSet<Q, _ResultSet<Q, Unit>>?>>,
	parameters: (ParametersSetter<Q>) -> Unit,
	data: Iterable<T>,
	rowFormatter: (T, ParametersSetter<InsertCreator.InsertParams.DataParam<Q>>) -> Unit,
	hasReturning: Boolean,
	returningInserted: (_ResultSet<Q, *>) -> Unit,
	returningUpdated: (_ResultSet<Q, *>) -> Unit,
	queryWithoutErrorMessage: String
) {
	contract {
		callsInPlace(queryGetter, InvocationKind.EXACTLY_ONCE)
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowFormatter, InvocationKind.UNKNOWN)
		callsInPlace(returningInserted, InvocationKind.AT_MOST_ONCE)
		callsInPlace(returningUpdated, InvocationKind.AT_MOST_ONCE)
	}
	val paramsSetter = queryGetter()

	_safeAutoClose(onAbort = { paramsSetter._abort() }) {
		parameters(paramsSetter)
	}

	val insertableValues = paramsSetter._execute()
	_safeAutoClose(onAbort = { insertableValues._abort() }) {
		for (e in data) {
			insertableValues._next()
			rowFormatter(e, insertableValues)
		}
	}

	val insertedRows = insertableValues._execute()
	if (insertedRows == null) {
		if (hasReturning)
			throw QueryWithoutReturnError(queryWithoutErrorMessage, compiledQuery)
		return
	}

	_safeAutoClose(onAbort = { insertedRows._abort() }) {
		returningInserted(insertedRows)
	}
	val updatedRows = insertedRows._close()
	_safeAutoClose(onAbort = { insertedRows._abort() }, onSuccess = { updatedRows._abort() }) {
		returningInserted(insertedRows)
	}
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, T> Transaction.insertParams(
	compiledQuery: Insert.InsertParams<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	data: Iterable<T>,
	rowFormatter: (T, ParametersSetter<InsertCreator.InsertParams.DataParam<QueryUserWrapper>>) -> Unit,
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowFormatter, InvocationKind.UNKNOWN)
	}
	_wrapInsertParamsWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._insertParams(compiledQuery) },
		parameters = parameters,
		data = data,
		rowFormatter = rowFormatter,
		hasReturning = false,
		returningInserted = {},
		returningUpdated = {},
		queryWithoutErrorMessage = ""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, T> SynchronizedDatabase<*>.autoTransactedInsertParams(
	compiledQuery: Insert.InsertParams<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	data: Iterable<T>,
	rowFormatter: (T, ParametersSetter<InsertCreator.InsertParams.DataParam<QueryUserWrapper>>) -> Unit,
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowFormatter, InvocationKind.UNKNOWN)
	}
	_wrapInsertParamsWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedInsertParams(compiledQuery) },
		parameters = parameters,
		data = data,
		rowFormatter = rowFormatter,
		hasReturning = false,
		returningInserted = {},
		returningUpdated = {},
		queryWithoutErrorMessage = ""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, T> Transaction.insertParamsReturningIterate(
	compiledQuery: Insert.InsertParams<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	data: Iterable<T>,
	rowFormatter: (T, ParametersSetter<InsertCreator.InsertParams.DataParam<QueryUserWrapper>>) -> Unit,
	insertedRows: (Row<QueryUserWrapper>) -> Unit,
	updatedRows: (Row<QueryUserWrapper>) -> Unit,
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowFormatter, InvocationKind.UNKNOWN)
	}
	_wrapInsertParamsWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._insertParams(compiledQuery) },
		parameters = parameters,
		data = data,
		rowFormatter = rowFormatter,
		hasReturning = true,
		returningInserted = { rs -> while (rs._next()) insertedRows(rs) },
		returningUpdated = { rs -> while (rs._next()) updatedRows(rs) },
		queryWithoutErrorMessage = ""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, T> SynchronizedDatabase<*>.autoTransactedInsertParametersReturningIterate(
	compiledQuery: Insert.InsertParams<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	data: Iterable<T>,
	rowFormatter: (T, ParametersSetter<InsertCreator.InsertParams.DataParam<QueryUserWrapper>>) -> Unit,
	insertedRows: (Row<QueryUserWrapper>) -> Unit,
	updatedRows: (Row<QueryUserWrapper>) -> Unit,
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowFormatter, InvocationKind.UNKNOWN)
	}
	_wrapInsertParamsWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedInsertParams(compiledQuery) },
		parameters = parameters,
		data = data,
		rowFormatter = rowFormatter,
		hasReturning = true,
		returningInserted = { rs -> while (rs._next()) insertedRows(rs) },
		returningUpdated = { rs -> while (rs._next()) updatedRows(rs) },
		queryWithoutErrorMessage = ""
	)
}


@PublishedApi
internal object SingleIterable : Iterable<Unit>, Iterator<Unit> {
	private var isReturned = false

	override fun iterator(): Iterator<Unit> = this

	override fun hasNext(): Boolean = !this.isReturned

	override fun next(): Unit {
		if (this.isReturned) throw IllegalStateException()
		this.isReturned = true
	}
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.insertSingle(
	compiledQuery: Insert.InsertParams<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowData: (ParametersSetter<InsertCreator.InsertParams.DataParam<QueryUserWrapper>>) -> Unit,
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowData, InvocationKind.EXACTLY_ONCE)
	}
	_wrapInsertParamsWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._insertParams(compiledQuery) },
		parameters = parameters,
		data = SingleIterable,
		rowFormatter = { _, r -> rowData(r) },
		hasReturning = false,
		returningInserted = {},
		returningUpdated = {},
		queryWithoutErrorMessage = ""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> SynchronizedDatabase<*>.autoTransactedInsertSingle(
	compiledQuery: Insert.InsertParams<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowData: (ParametersSetter<InsertCreator.InsertParams.DataParam<QueryUserWrapper>>) -> Unit,
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowData, InvocationKind.EXACTLY_ONCE)
	}
	_wrapInsertParamsWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedInsertParams(compiledQuery) },
		parameters = parameters,
		data = SingleIterable,
		rowFormatter = { _, r -> rowData(r) },
		hasReturning = false,
		returningInserted = {},
		returningUpdated = {},
		queryWithoutErrorMessage = ""
	)
}

@Suppress("DuplicatedCode")
@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.insertSingleReturning(
	compiledQuery: Insert.InsertParams<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowData: (ParametersSetter<InsertCreator.InsertParams.DataParam<QueryUserWrapper>>) -> Unit = {},
	insertedRows: (Row<QueryUserWrapper>) -> R,
	updatedRows: (Row<QueryUserWrapper>) -> R
): R {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowData, InvocationKind.EXACTLY_ONCE)
		callsInPlace(insertedRows, InvocationKind.AT_MOST_ONCE)
		callsInPlace(updatedRows, InvocationKind.AT_MOST_ONCE)
	}
	var isReturned = false
	var ret: R? = null
	_wrapInsertParamsWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._insertParams(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		data = SingleIterable,
		rowFormatter = { _, r -> rowData(r) },
		returningInserted = returningInserted@{ rs ->
			if (!rs._next()) return@returningInserted
			isReturned = true
			ret = insertedRows(rs)
			if (rs._next()) throw ResultNotSingleError()
		},
		returningUpdated = returningUpdated@{ rs ->
			if (!rs._next()) return@returningUpdated
			if (isReturned) throw ResultNotSingleError()
			isReturned = true
			ret = insertedRows(rs)
			if (rs._next()) throw ResultNotSingleError()
		},
		queryWithoutErrorMessage = ""
	)
	if (!isReturned)
		throw QueryWithoutReturnError("", compiledQuery)
	@Suppress("UNCHECKED_CAST")
	return ret as R
}

@Suppress("DuplicatedCode")
@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> SynchronizedDatabase<*>.autoTransactedInsertSingleReturning(
	compiledQuery: Insert.InsertParams<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowData: (ParametersSetter<InsertCreator.InsertParams.DataParam<QueryUserWrapper>>) -> Unit = {},
	insertedRows: (Row<QueryUserWrapper>) -> R,
	updatedRows: (Row<QueryUserWrapper>) -> R
): R {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowData, InvocationKind.EXACTLY_ONCE)
		callsInPlace(insertedRows, InvocationKind.AT_MOST_ONCE)
		callsInPlace(updatedRows, InvocationKind.AT_MOST_ONCE)
	}
	var isReturned = false
	var ret: R? = null
	_wrapInsertParamsWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedInsertParams(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		data = SingleIterable,
		rowFormatter = { _, r -> rowData(r) },
		returningInserted = returningInserted@{ rs ->
			if (!rs._next()) return@returningInserted
			isReturned = true
			ret = insertedRows(rs)
			if (rs._next()) throw ResultNotSingleError()
		},
		returningUpdated = returningUpdated@{ rs ->
			if (!rs._next()) return@returningUpdated
			if (isReturned) throw ResultNotSingleError()
			isReturned = true
			ret = insertedRows(rs)
			if (rs._next()) throw ResultNotSingleError()
		},
		queryWithoutErrorMessage = ""
	)
	if (!isReturned)
		throw QueryWithoutReturnError("", compiledQuery)
	@Suppress("UNCHECKED_CAST")
	return ret as R
}

@Suppress("FunctionName", "DuplicatedCode")
@PublishedApi
@OptIn(LowLevelApi::class)
internal suspend inline fun <Q : Any> _wrapInsertFromQueryWithLambdas(
	compiledQuery: Insert.InsertFromQuery<Q>,
	queryGetter: () -> _ParametersSetter<Q, _ResultSet<Q, _ResultSet<Q, Unit>>?>,
	parameters: (ParametersSetter<Q>) -> Unit,
	hasReturning: Boolean,
	returningInserted: (_ResultSet<Q, *>) -> Unit,
	returningUpdated: (_ResultSet<Q, *>) -> Unit,
	queryWithoutErrorMessage: String
) {
	contract {
		callsInPlace(queryGetter, InvocationKind.EXACTLY_ONCE)
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(returningInserted, InvocationKind.AT_MOST_ONCE)
		callsInPlace(returningUpdated, InvocationKind.AT_MOST_ONCE)
	}
	val paramsSetter = queryGetter()

	_safeAutoClose(onAbort = { paramsSetter._abort() }) {
		parameters(paramsSetter)
	}

	val insertedRows = paramsSetter._execute()
	if (insertedRows == null) {
		if (hasReturning)
			throw QueryWithoutReturnError(queryWithoutErrorMessage, compiledQuery)
		return
	}

	_safeAutoClose(onAbort = { insertedRows._abort() }) {
		returningInserted(insertedRows)
	}
	val updatedRows = insertedRows._close()
	_safeAutoClose(onAbort = { insertedRows._abort() }, onSuccess = { updatedRows._abort() }) {
		returningInserted(insertedRows)
	}
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.insertFromQuery(
	compiledQuery: Insert.InsertFromQuery<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
	}
	_wrapInsertFromQueryWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._insertFromQuery(compiledQuery) },
		parameters = parameters,
		hasReturning = false,
		returningInserted = {},
		returningUpdated = {},
		queryWithoutErrorMessage = ""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> SynchronizedDatabase<*>.autoTransactedInsertFromQuery(
	compiledQuery: Insert.InsertFromQuery<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
	}
	_wrapInsertFromQueryWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedInsertFromQuery(compiledQuery) },
		parameters = parameters,
		hasReturning = false,
		returningInserted = {},
		returningUpdated = {},
		queryWithoutErrorMessage = ""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.insertFromQueryReturningIterate(
	compiledQuery: Insert.InsertFromQuery<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	insertedRows: (Row<QueryUserWrapper>) -> Unit,
	updatedRows: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(insertedRows, InvocationKind.UNKNOWN)
		callsInPlace(updatedRows, InvocationKind.UNKNOWN)
	}
	_wrapInsertFromQueryWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._insertFromQuery(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returningInserted = { rs -> while (rs._next()) insertedRows(rs) },
		returningUpdated = { rs -> while (rs._next()) updatedRows(rs) },
		queryWithoutErrorMessage = ""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> SynchronizedDatabase<*>.autoTransactedInsertFromQueryReturningIterate(
	compiledQuery: Insert.InsertFromQuery<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	insertedRows: (Row<QueryUserWrapper>) -> Unit,
	updatedRows: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(insertedRows, InvocationKind.UNKNOWN)
		callsInPlace(updatedRows, InvocationKind.UNKNOWN)
	}
	_wrapInsertFromQueryWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedInsertFromQuery(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returningInserted = { rs -> while (rs._next()) insertedRows(rs) },
		returningUpdated = { rs -> while (rs._next()) updatedRows(rs) },
		queryWithoutErrorMessage = ""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.insertFromQueryReturningSingleOrNull(
	compiledQuery: Insert.InsertFromQuery<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	insertedRows: (Row<QueryUserWrapper>) -> R,
	updatedRows: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(insertedRows, InvocationKind.AT_MOST_ONCE)
		callsInPlace(updatedRows, InvocationKind.AT_MOST_ONCE)
	}
	var isReturned = false
	var ret: R? = null
	_wrapInsertFromQueryWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._insertFromQuery(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returningInserted = returningInserted@{ rs ->
			if (!rs._next()) return@returningInserted
			isReturned = true
			ret = insertedRows(rs)
			if (rs._next()) throw ResultNotSingleError()
		},
		returningUpdated = returningUpdated@{ rs ->
			if (!rs._next()) return@returningUpdated
			if (isReturned) throw ResultNotSingleError()
			isReturned = true
			ret = insertedRows(rs)
			if (rs._next()) throw ResultNotSingleError()
		},
		queryWithoutErrorMessage = ""
	)
	if (!isReturned)
		return null
	return ret
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> SynchronizedDatabase<*>.autoTransactedInsertFromQueryReturningSingleOrNull(
	compiledQuery: Insert.InsertFromQuery<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	insertedRows: (Row<QueryUserWrapper>) -> R,
	updatedRows: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(insertedRows, InvocationKind.AT_MOST_ONCE)
		callsInPlace(updatedRows, InvocationKind.AT_MOST_ONCE)
	}
	var isReturned = false
	var ret: R? = null
	_wrapInsertFromQueryWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedInsertFromQuery(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returningInserted = returningInserted@{ rs ->
			if (!rs._next()) return@returningInserted
			isReturned = true
			ret = insertedRows(rs)
			if (rs._next()) throw ResultNotSingleError()
		},
		returningUpdated = returningUpdated@{ rs ->
			if (!rs._next()) return@returningUpdated
			if (isReturned) throw ResultNotSingleError()
			isReturned = true
			ret = insertedRows(rs)
			if (rs._next()) throw ResultNotSingleError()
		},
		queryWithoutErrorMessage = ""
	)
	if (!isReturned)
		return null
	return ret
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.update(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
	}
	_wrapWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._update(compiledQuery) },
		parameters = parameters,
		hasReturning = false,
		returning = {},
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> SynchronizedDatabase<*>.autoTransactedUpdate(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
	}
	_wrapWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedUpdate(compiledQuery) },
		parameters = parameters,
		hasReturning = false,
		returning = {},
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.updateReturningSingleOrNull(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.AT_MOST_ONCE)
	}
	return _wrapWithLambdasSingleOrNull(
		compiledQuery = compiledQuery,
		queryGetter = { this._update(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}


@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> SynchronizedDatabase<*>.autoTransactedUpdateReturningSingleOrNull(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.AT_MOST_ONCE)
	}
	return _wrapWithLambdasSingleOrNull(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedUpdate(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.updateReturningIterate(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	_wrapWithLambdasIterate(
		compiledQuery = compiledQuery,
		queryGetter = { this._update(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> SynchronizedDatabase<*>.autoTransactedUpdateReturningIterate(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	_wrapWithLambdasIterate(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedUpdate(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.updateReturningMap(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): List<R> {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	return _wrapWithLambdasMap(
		compiledQuery = compiledQuery,
		queryGetter = { this._update(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> SynchronizedDatabase<*>.autoTransactedUpdateReturningMap(
	compiledQuery: Update<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): List<R> {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	return _wrapWithLambdasMap(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedUpdate(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.delete(
	compiledQuery: Delete<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
	}
	_wrapWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._delete(compiledQuery) },
		parameters = parameters,
		hasReturning = false,
		returning = {},
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> SynchronizedDatabase<*>.autoTransactedDelete(
	compiledQuery: Delete<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
	}
	_wrapWithLambdas(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedDelete(compiledQuery) },
		parameters = parameters,
		hasReturning = false,
		returning = {},
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.deleteReturningSingleOrNull(
	compiledQuery: Delete<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.AT_MOST_ONCE)
	}
	return _wrapWithLambdasSingleOrNull(
		compiledQuery = compiledQuery,
		queryGetter = { this._delete(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}


@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> SynchronizedDatabase<*>.autoTransactedDeleteReturningSingleOrNull(
	compiledQuery: Delete<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit = {},
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): R? {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.AT_MOST_ONCE)
	}
	return _wrapWithLambdasSingleOrNull(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedDelete(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> Transaction.deleteReturningIterate(
	compiledQuery: Delete<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	_wrapWithLambdasIterate(
		compiledQuery = compiledQuery,
		queryGetter = { this._delete(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any> SynchronizedDatabase<*>.autoTransactedDeleteReturningIterate(
	compiledQuery: Delete<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> Unit
) {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	_wrapWithLambdasIterate(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedDelete(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> Transaction.deleteReturningMap(
	compiledQuery: Delete<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): List<R> {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	return _wrapWithLambdasMap(
		compiledQuery = compiledQuery,
		queryGetter = { this._delete(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend inline fun <QueryUserWrapper : Any, R> SynchronizedDatabase<*>.autoTransactedDeleteReturningMap(
	compiledQuery: Delete<QueryUserWrapper>,
	parameters: (ParametersSetter<QueryUserWrapper>) -> Unit,
	rowsConsumer: (Row<QueryUserWrapper>) -> R
): List<R> {
	contract {
		callsInPlace(parameters, InvocationKind.EXACTLY_ONCE)
		callsInPlace(rowsConsumer, InvocationKind.UNKNOWN)
	}
	return _wrapWithLambdasMap(
		compiledQuery = compiledQuery,
		queryGetter = { this._autoTransactedDelete(compiledQuery) },
		parameters = parameters,
		hasReturning = true,
		returning = rowsConsumer,
		""
	)
}

@OptIn(LowLevelApi::class)
public suspend fun <R> SynchronizedDatabase<*>.transaction(body: (Transaction) -> R): R {
	contract {
		callsInPlace(body, InvocationKind.EXACTLY_ONCE)
	}

	val tr = this._startTransaction()
	_safeAutoClose(onAbort = { tr.rollback(); tr._assertTransactionFinishedAndReleaseResources() }, onSuccess = { tr._assertTransactionFinishedAndReleaseResources() }) {
		return@transaction body(tr)
	}
}